@@ -32,11 +32,13 @@ CONF = cfg.CONF
 LOG = log.getLogger(__name__)
 
 
-# TODO(gmann): Remove setting the default value of config policy_file
-# once oslo_policy change the default value to 'policy.yaml'.
-# https://github.com/openstack/oslo.policy/blob/a626ad12fe5a3abd49d70e3e5b95589d279ab578/oslo_policy/opts.py#L49
+# TODO(gmann): Remove overriding the default value of config options
+# 'policy_file' once oslo_policy change its default value to what
+# is overridden here.
 DEFAULT_POLICY_FILE = 'policy.yaml'
-opts.set_defaults(cfg.CONF, DEFAULT_POLICY_FILE)
+opts.set_defaults(
+    cfg.CONF,
+    DEFAULT_POLICY_FILE)
 
 # Generic policy check string for system administrators. These are the people
 # who need the highest level of authorization to operate the deployment.
@@ -47,17 +49,24 @@ SYSTEM_ADMIN = 'role:admin and system_scope:all'
 
 # Generic policy check string for system users who don't require all the
 # authorization that system administrators typically have. This persona, or
-# check string, typically isn't used by default, but it's existence it useful
+# check string, typically isn't used by default, but it's existence is useful
 # in the event a deployment wants to offload some administrative action from
-# system administrator to system members
-SYSTEM_MEMBER = 'role:member and system_scope:all'
+# system administrator to system members.
+# The rule:service_role match here is to enable an elevated level of API
+# access for a specialized service role and users with appropriate
+# service role access.
+SYSTEM_MEMBER = '(role:member and system_scope:all) or rule:service_role'  # noqa
 
-# Generic policy check string for read-only access to system-level resources.
-# This persona is useful for someone who needs access for auditing or even
-# support. These uses are also able to view project-specific resources where
-# applicable (e.g., listing all volumes in the deployment, regardless of the
-# project they belong to).
-SYSTEM_READER = 'role:reader and system_scope:all'
+# Generic policy check string for read-only access to system-level
+# resources. This persona is useful for someone who needs access
+# for auditing or even support. These users are also able to view
+# project-specific resources where applicable (e.g., listing all
+# volumes in the deployment, regardless of the project they belong to).
+# The rule:service_role match here is to enable an elevated level of API
+# access for a specialized service role and users with appropriate
+# role access, specifically because 'service" role is outside of the RBAC
+# model defaults and does not imply reader access.
+SYSTEM_READER = '(role:reader and system_scope:all) or (role:service and system_scope:all) or rule:service_role'  # noqa
 
 # This check string is reserved for actions that require the highest level of
 # authorization on a project or resources within the project (e.g., setting the
@@ -83,15 +92,23 @@ PROJECT_MEMBER = ('role:member and '
 PROJECT_READER = ('role:reader and '
                   '(project_id:%(node.owner)s or project_id:%(node.lessee)s)')
 
+# This check string is used for granting access to other services which need
+# to communicate with Ironic, for example, Nova-Compute to provision nodes,
+# or Ironic-Inspector to create nodes. The idea behind a service role is
+# one which has restricted access to perform operations, that are limited
+# to remote automated and inter-operation processes.
+SYSTEM_SERVICE = ('role:service and system_scope:all')
+PROJECT_SERVICE = ('role:service and project_id:%(node.owner)s')
+
 # The following are common composite check strings that are useful for
 # protecting APIs designed to operate with multiple scopes (e.g., a system
 # administrator should be able to delete any baremetal host in the deployment,
 # a project member should only be able to delete hosts in their project).
 SYSTEM_OR_PROJECT_MEMBER = (
-    '(' + SYSTEM_MEMBER + ') or (' + PROJECT_MEMBER + ')'
+    '(' + SYSTEM_MEMBER + ') or (' + PROJECT_MEMBER + ') or (' + SYSTEM_SERVICE + ')'  # noqa
 )
 SYSTEM_OR_PROJECT_READER = (
-    '(' + SYSTEM_READER + ') or (' + PROJECT_READER + ')'
+    '(' + SYSTEM_READER + ') or (' + PROJECT_READER + ') or (' + PROJECT_SERVICE + ')'  # noqa
 )
 
 PROJECT_OWNER_ADMIN = ('role:admin and project_id:%(node.owner)s')
@@ -109,28 +126,54 @@ ALLOCATION_OWNER_MANAGER = ('role:manager and project_id:%(allocation.owner)s')
 ALLOCATION_OWNER_MEMBER = ('role:member and project_id:%(allocation.owner)s')
 ALLOCATION_OWNER_READER = ('role:reader and project_id:%(allocation.owner)s')
 
+# Members can create/destroy their runbooks.
+RUNBOOK_OWNER_ADMIN = ('role:admin and project_id:%(runbook.owner)s')
+RUNBOOK_OWNER_MANAGER = ('role:manager and project_id:%(runbook.owner)s')
+RUNBOOK_OWNER_MEMBER = ('role:member and project_id:%(runbook.owner)s')
+RUNBOOK_OWNER_READER = ('role:reader and project_id:%(runbook.owner)s')
+
+RUNBOOK_ADMIN = (
+    '(' + SYSTEM_MEMBER + ') or (' + RUNBOOK_OWNER_MANAGER + ') or role:service' # noqa
+)
+
+RUNBOOK_READER = (
+    '(' + SYSTEM_READER + ') or (' + RUNBOOK_OWNER_READER + ') or role:service' # noqa
+)
+
+RUNBOOK_CREATOR = (
+    '(' + SYSTEM_MEMBER + ') or role:manager or role:service' # noqa
+)
+
+# Used for general operations like changing provision state.
 SYSTEM_OR_OWNER_MEMBER_AND_LESSEE_ADMIN = (
-    '(' + SYSTEM_MEMBER + ') or (' + PROJECT_OWNER_MEMBER + ') or (' + PROJECT_LESSEE_ADMIN + ') or (' + PROJECT_LESSEE_MANAGER + ')'  # noqa
+    '(' + SYSTEM_MEMBER + ') or (' + SYSTEM_SERVICE + ') or (' + PROJECT_OWNER_MEMBER + ') or (' + PROJECT_LESSEE_ADMIN + ') or (' + PROJECT_LESSEE_MANAGER + ') or (' + PROJECT_SERVICE + ')'  # noqa
 )
 
+# Used for creation and deletion of network ports.
 SYSTEM_ADMIN_OR_OWNER_ADMIN = (
-    '(' + SYSTEM_ADMIN + ') or (' + PROJECT_OWNER_ADMIN + ') or (' + PROJECT_OWNER_MANAGER + ')'  # noqa
+    '(' + SYSTEM_ADMIN + ') or (' + SYSTEM_SERVICE + ') or (' + PROJECT_OWNER_ADMIN + ') or (' + PROJECT_OWNER_MANAGER + ') or (' + PROJECT_SERVICE + ')'  # noqa
 )
 
+# Used to map system members, and owner admins to the same access rights.
+# This is actions such as update driver interfaces, delete ports.
 SYSTEM_MEMBER_OR_OWNER_ADMIN = (
-    '(' + SYSTEM_MEMBER + ') or (' + PROJECT_OWNER_ADMIN + ') or (' + PROJECT_OWNER_MANAGER + ')'  # noqa
+    '(' + SYSTEM_MEMBER + ') or (' + SYSTEM_SERVICE + ') or (' + PROJECT_OWNER_ADMIN + ') or (' + PROJECT_OWNER_MANAGER + ') or (' + PROJECT_SERVICE + ')'  # noqa
 )
 
+# Used to map "member" only rights, i.e. those of "users using a deployment"
 SYSTEM_MEMBER_OR_OWNER_MEMBER = (
-    '(' + SYSTEM_MEMBER + ') or (' + PROJECT_OWNER_MEMBER + ')'
+    '(' + SYSTEM_MEMBER + ') or (' + SYSTEM_SERVICE + ') or (' + PROJECT_OWNER_MEMBER + ') or (' + PROJECT_SERVICE + ')'  # noqa
 )
 
+# Used throughout to map where authenticated readers
+# should be able to read API objects.
 SYSTEM_OR_OWNER_READER = (
-    '(' + SYSTEM_READER + ') or (' + PROJECT_OWNER_READER + ')'
+    '(' + SYSTEM_READER + ') or (' + SYSTEM_SERVICE + ') or (' + PROJECT_OWNER_READER + ') or (' + PROJECT_SERVICE + ')'  # noqa
 )
 
+# Mainly used for targets/connectors
 SYSTEM_MEMBER_OR_OWNER_LESSEE_ADMIN = (
-    '(' + SYSTEM_MEMBER + ') or (' + PROJECT_OWNER_ADMIN + ') or (' + PROJECT_OWNER_MANAGER + ') or (' + PROJECT_LESSEE_ADMIN + ') or (' + PROJECT_LESSEE_MANAGER + ')'   # noqa
+    '(' + SYSTEM_MEMBER + ') or (' + SYSTEM_SERVICE + ') or (' + PROJECT_OWNER_ADMIN + ') or (' + PROJECT_OWNER_MANAGER + ') or (' + PROJECT_LESSEE_ADMIN + ') or (' + PROJECT_LESSEE_MANAGER + ') or (' + PROJECT_SERVICE + ')'  # noqa
 )
 
 
@@ -152,7 +195,10 @@ ALLOCATION_CREATOR = (
 # Special purpose aliases for things like "ability to access the API
 # as a reader, or permission checking that does not require node
 # owner relationship checking
-API_READER = ('role:reader')
+API_READER = ('(role:reader) or (role:service)')
+
+# Used for ability to view target properties of a volume, which is
+# considered highly restricted.
 TARGET_PROPERTIES_READER = (
     '(' + SYSTEM_READER + ') or (role:admin)'
 )
@@ -184,11 +230,18 @@ default_policies = [
     #             the value of this field
     policy.RuleDefault('show_password',
                        '!',
-                       description='Show or mask secrets within node driver information in API responses'),  # noqa
+                       description='Show or mask secrets within node driver information in API responses. This setting should be used with the utmost care as its use can present a security risk.'),  # noqa
     # Generic default to hide instance secrets
     policy.RuleDefault('show_instance_secrets',
                        '!',
-                       description='Show or mask secrets within instance information in API responses'),  # noqa
+                       description='Show or mask secrets within instance information in API responses. This setting should be used with the utmost care as its use can present a security risk.'),  # noqa
+    # NOTE(TheJulia): This is a special rule to allow customization of the
+    # service role check. The config.service_project_name is a reserved
+    # target check field which is loaded from configuration to the
+    # check context in ironic/common/context.py.
+    policy.RuleDefault('service_role',
+                       'role:service and project_name:%(config.service_project_name)s',  # noqa
+                       description='Rule to match service role usage with a service project, delineated as a separate rule to enable customization.'),  # noqa
     # Roles likely to be overridden by operator
     # TODO(TheJulia): Lets nuke demo from high orbit.
     policy.RuleDefault('is_member',
@@ -436,7 +489,7 @@ deprecated_bios_disable_cleaning = policy.DeprecatedRule(
 node_policies = [
     policy.DocumentedRuleDefault(
         name='baremetal:node:create',
-        check_str=SYSTEM_ADMIN,
+        check_str='(' + SYSTEM_ADMIN + ') or (' + SYSTEM_SERVICE + ')',
         scope_types=['system', 'project'],
         description='Create Node records',
         operations=[{'path': '/nodes', 'method': 'POST'}],
@@ -444,8 +497,8 @@ node_policies = [
     ),
     policy.DocumentedRuleDefault(
         name='baremetal:node:create:self_owned_node',
-        check_str=('role:admin'),
-        scope_types=['project'],
+        check_str=('(role:admin) or (role:service)'),
+        scope_types=['system', 'project'],
         description='Create node records which will be tracked '
                     'as owned by the associated user project.',
         operations=[{'path': '/nodes', 'method': 'POST'}],
@@ -463,7 +516,7 @@ node_policies = [
     policy.DocumentedRuleDefault(
         name='baremetal:node:list_all',
         check_str=SYSTEM_READER,
-        scope_types=['system'],
+        scope_types=['system', 'project'],
         description='Retrieve multiple Node records',
         operations=[{'path': '/nodes', 'method': 'GET'},
                     {'path': '/nodes/detail', 'method': 'GET'}],
@@ -674,7 +727,7 @@ node_policies = [
     policy.DocumentedRuleDefault(
         name='baremetal:node:delete:self_owned_node',
         check_str=PROJECT_ADMIN,
-        scope_types=['project'],
+        scope_types=['system', 'project'],
         description='Delete node records which are associated with '
                     'the requesting project.',
         operations=[{'path': '/nodes/{node_ident}', 'method': 'DELETE'}],
@@ -713,7 +766,7 @@ node_policies = [
         deprecated_rule=deprecated_node_clear_maintenance
     ),
 
-    # NOTE(TheJulia): This should liekly be deprecated and be replaced with
+    # NOTE(TheJulia): This should likely be deprecated and be replaced with
     # a cached object.
     policy.DocumentedRuleDefault(
         name='baremetal:node:get_boot_device',
@@ -827,6 +880,24 @@ node_policies = [
         ],
         deprecated_rule=deprecated_node_set_provision_state
     ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:set_provision_state:clean_steps',
+        check_str=SYSTEM_OR_OWNER_MEMBER_AND_LESSEE_ADMIN,
+        scope_types=['system', 'project'],
+        description='Allow execution of arbitrary steps on a node',
+        operations=[
+            {'path': '/nodes/{node_ident}/states/provision', 'method': 'PUT'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:set_provision_state:service_steps',
+        check_str=SYSTEM_OR_OWNER_MEMBER_AND_LESSEE_ADMIN,
+        scope_types=['system', 'project'],
+        description='Allow execution of arbitrary steps on a node',
+        operations=[
+            {'path': '/nodes/{node_ident}/states/provision', 'method': 'PUT'}
+        ],
+    ),
     policy.DocumentedRuleDefault(
         name='baremetal:node:set_raid_state',
         check_str=SYSTEM_MEMBER_OR_OWNER_MEMBER,
@@ -942,7 +1013,7 @@ node_policies = [
         name='baremetal:node:history:get',
         check_str=SYSTEM_OR_OWNER_READER,
         scope_types=['system', 'project'],
-        description='Filter to allow operators to retreive history records '
+        description='Filter to allow operators to retrieve history records '
                     'for a node.',
         operations=[
             {'path': '/nodes/{node_ident}/history', 'method': 'GET'},
@@ -954,8 +1025,78 @@ node_policies = [
         # operating context.
         deprecated_rule=deprecated_node_get
     ),
-
-
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:inventory:get',
+        check_str=SYSTEM_OR_OWNER_READER,
+        scope_types=['system', 'project'],
+        description='Retrieve introspection data for a node.',
+        operations=[
+            {'path': '/nodes/{node_ident}/inventory', 'method': 'GET'},
+        ],
+        # This rule fallsback to deprecated_node_get in order to provide a
+        # mechanism so the additional policies only engage in an updated
+        # operating context.
+        deprecated_rule=deprecated_node_get
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:update:shard',
+        check_str=SYSTEM_ADMIN,
+        scope_types=['system', 'project'],
+        description='Governs if node shard field can be updated via '
+                    'the API clients.',
+        operations=[{'path': '/nodes/{node_ident}', 'method': 'PATCH'}],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:shards:get',
+        check_str=SYSTEM_READER,
+        scope_types=['system', 'project'],
+        description='Governs if shards can be read via the API clients.',
+        operations=[{'path': '/shards', 'method': 'GET'}],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:update:parent_node',
+        check_str=SYSTEM_MEMBER,
+        scope_types=['system', 'project'],
+        description='Governs if node parent_node field can be updated via '
+                    'the API clients.',
+        operations=[{'path': '/nodes/{node_ident}', 'method': 'PATCH'}],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:firmware:get',
+        check_str=SYSTEM_OR_PROJECT_READER,
+        scope_types=['system', 'project'],
+        description='Retrieve Node Firmware components information',
+        operations=[
+            {'path': '/nodes/{node_ident}/firmware', 'method': 'GET'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:vmedia:attach',
+        check_str=SYSTEM_OR_PROJECT_MEMBER,
+        scope_types=['system', 'project'],
+        description='Attach a virtual media device to a node',
+        operations=[
+            {'path': '/nodes/{node_ident}/vmedia', 'method': 'POST'}\
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:vmedia:detach',
+        check_str=SYSTEM_OR_PROJECT_MEMBER,
+        scope_types=['system', 'project'],
+        description='Detach a virtual media device from a node',
+        operations=[
+            {'path': '/nodes/{node_ident}/vmedia', 'method': 'DELETE'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:node:vmedia:get',
+        check_str=SYSTEM_OR_PROJECT_READER,
+        scope_types=['system', 'project'],
+        description='Get virtual media device details from a node',
+        operations=[
+            {'path': '/nodes/{node_ident}/vmedia', 'method': 'GET'}
+        ],
+    ),
 ]
 
 deprecated_port_reason = """
@@ -1390,6 +1531,12 @@ utility_policies = [
         operations=[{'path': '/lookup', 'method': 'GET'}],
         deprecated_rule=deprecated_ipa_lookup
     ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:driver:ipa_continue_inspection',
+        check_str='',
+        description='Receive inspection data from the ramdisk',
+        operations=[{'path': '/continue_inspection', 'method': 'POST'}],
+    ),
 ]
 
 
@@ -1536,7 +1683,7 @@ conductor_policies = [
     policy.DocumentedRuleDefault(
         name='baremetal:conductor:get',
         check_str=SYSTEM_READER,
-        scope_types=['system'],
+        scope_types=['system', 'project'],
         description='Retrieve Conductor records',
         operations=[
             {'path': '/conductors', 'method': 'GET'},
@@ -1667,7 +1814,7 @@ allocation_policies = [
         # The latter is more for projects and services using admin project
         # rights. Specific checking because of the expanded rights of
         # this functionality.
-        check_str=('(rule:is_member and role:baremetal_admin) or (is_admin_project:True and role:admin)'),  # noqa 
+        check_str=('(rule:is_member and role:baremetal_admin) or (is_admin_project:True and role:admin)'),  # noqa
         scope_types=['project'],
         description=('Logical restrictor to prevent legacy allocation rule '
                      'missuse - Requires blank allocations to originate from '
@@ -1737,7 +1884,7 @@ deploy_template_policies = [
     policy.DocumentedRuleDefault(
         name='baremetal:deploy_template:get',
         check_str=SYSTEM_READER,
-        scope_types=['system'],
+        scope_types=['system', 'project'],
         description='Retrieve Deploy Template records',
         operations=[
             {'path': '/deploy_templates', 'method': 'GET'},
@@ -1749,7 +1896,7 @@ deploy_template_policies = [
     policy.DocumentedRuleDefault(
         name='baremetal:deploy_template:create',
         check_str=SYSTEM_ADMIN,
-        scope_types=['system'],
+        scope_types=['system', 'project'],
         description='Create Deploy Template records',
         operations=[{'path': '/deploy_templates', 'method': 'POST'}],
         deprecated_rule=deprecated_deploy_template_create
@@ -1757,7 +1904,7 @@ deploy_template_policies = [
     policy.DocumentedRuleDefault(
         name='baremetal:deploy_template:delete',
         check_str=SYSTEM_ADMIN,
-        scope_types=['system'],
+        scope_types=['system', 'project'],
         description='Delete Deploy Template records',
         operations=[
             {'path': '/deploy_templates/{deploy_template_ident}',
@@ -1768,7 +1915,7 @@ deploy_template_policies = [
     policy.DocumentedRuleDefault(
         name='baremetal:deploy_template:update',
         check_str=SYSTEM_ADMIN,
-        scope_types=['system'],
+        scope_types=['system', 'project'],
         description='Update Deploy Template records',
         operations=[
             {'path': '/deploy_templates/{deploy_template_ident}',
@@ -1778,6 +1925,89 @@ deploy_template_policies = [
     ),
 ]
 
+runbook_policies = [
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:get',
+        check_str=RUNBOOK_READER,
+        scope_types=['system', 'project'],
+        description='Retrieve a single runbook record',
+        operations=[
+            {'path': '/runbooks/{runbook_ident}', 'method': 'GET'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:list',
+        check_str=API_READER,
+        scope_types=['system', 'project'],
+        description='Retrieve multiple runbook records, filtered by '
+                    'an explicit owner or the client project_id',
+        operations=[
+            {'path': '/runbooks', 'method': 'GET'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:list_all',
+        check_str=SYSTEM_READER,
+        scope_types=['system', 'project'],
+        description='Retrieve all runbook records',
+        operations=[
+            {'path': '/runbooks', 'method': 'GET'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:create',
+        check_str=RUNBOOK_CREATOR,
+        scope_types=['system', 'project'],
+        description='Create Runbook records',
+        operations=[{'path': '/runbooks', 'method': 'POST'}],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:delete',
+        check_str=RUNBOOK_ADMIN,
+        scope_types=['system', 'project'],
+        description='Delete a runbook record',
+        operations=[
+            {'path': '/runbooks/{runbook_ident}', 'method': 'DELETE'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:update',
+        check_str=RUNBOOK_ADMIN,
+        scope_types=['system', 'project'],
+        description='Update a runbook record',
+        operations=[
+            {'path': '/runbooks/{runbook_ident}', 'method': 'PATCH'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:update:public',
+        check_str=SYSTEM_MEMBER,
+        scope_types=['system', 'project'],
+        description='Set and unset a runbook as public',
+        operations=[
+            {'path': '/runbooks/{runbook_ident}/public', 'method': 'PATCH'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:update:owner',
+        check_str=SYSTEM_MEMBER,
+        scope_types=['system', 'project'],
+        description='Set and unset the owner of a runbook',
+        operations=[
+            {'path': '/runbooks/{runbook_ident}/owner', 'method': 'PATCH'}
+        ],
+    ),
+    policy.DocumentedRuleDefault(
+        name='baremetal:runbook:use',
+        check_str=RUNBOOK_ADMIN,
+        scope_types=['system', 'project'],
+        description='Allowed to use a runbook for node operations',
+        operations=[
+            {'path': '/nodes/{node_ident}/states/provision', 'method': 'PUT'}
+        ],
+    )
+]
+
 
 def list_policies():
     policies = itertools.chain(
@@ -1794,6 +2024,7 @@ def list_policies():
         allocation_policies,
         event_policies,
         deploy_template_policies,
+        runbook_policies,
     )
     return policies
 
@@ -1827,7 +2058,7 @@ def init_enforcer(policy_file=None, rules=None,
         rules=rules,
         default_rule=default_rule,
         use_conf=use_conf)
-    # NOTE(melwitt): Explictly disable the warnings for policies
+    # NOTE(melwitt): Explicitly disable the warnings for policies
     # changing their default check_str. During policy-defaults-refresh
     # work, all the policy defaults have been changed and warning for
     # each policy started filling the logs limit for various tool.
