@@ -18,16 +18,25 @@
 from oslo_config import cfg
 from oslo_config import types
 
+from ironic.common import boot_modes
 from ironic.common.i18n import _
+from ironic.common import lessee_sources
+
 
 opts = [
     cfg.IntOpt('workers_pool_size',
-               default=100, min=3,
+               default=300, min=3,
                help=_('The size of the workers greenthread pool. '
                       'Note that 2 threads will be reserved by the conductor '
                       'itself for handling heart beats and periodic tasks. '
                       'On top of that, `sync_power_state_workers` will take '
                       'up to 7 green threads with the default value of 8.')),
+    cfg.IntOpt('reserved_workers_pool_percentage',
+               default=5, min=0, max=50,
+               help=_('The percentage of the whole workers pool that will be '
+                      'kept for API requests and other important tasks. '
+                      'This part of the pool will not be used for periodic '
+                      'tasks or agent heartbeats. Set to 0 to disable.')),
     cfg.IntOpt('heartbeat_interval',
                default=10,
                help=_('Seconds between conductor heart beats.')),
@@ -205,17 +214,11 @@ opts = [
                        'endpoint via multicast DNS.')),
     cfg.StrOpt('deploy_kernel',
                mutable=True,
-               deprecated_for_removal=True,
-               deprecated_reason=_('Replaced by deploy_kernel_by_arch which '
-                                   'provides more configuration options.'),
-               help=_('DEPRECATED: Glance ID, http:// or file:// URL of the '
+               help=_('Glance ID, http:// or file:// URL of the '
                       'kernel of the default deploy image.')),
     cfg.StrOpt('deploy_ramdisk',
                mutable=True,
-               deprecated_for_removal=True,
-               deprecated_reason=_('Replaced by deploy_ramdisk_by_arch which '
-                                   'provides more configuration options.'),
-               help=_('DEPRECATED: Glance ID, http:// or file:// URL of the '
+               help=_('Glance ID, http:// or file:// URL of the '
                       'initramfs of the default deploy image.')),
     cfg.DictOpt('deploy_kernel_by_arch',
                 default={},
@@ -231,17 +234,11 @@ opts = [
                        'initramfs of the default deploy image.')),
     cfg.StrOpt('rescue_kernel',
                mutable=True,
-               deprecated_for_removal=True,
-               deprecated_reason=_('Replaced by rescue_kernel_by_arch which '
-                                   'provides more configuration options.'),
-               help=_('DEPRECATED: Glance ID, http:// or file:// URL of the '
+               help=_('Glance ID, http:// or file:// URL of the '
                       'kernel of the default rescue image.')),
     cfg.StrOpt('rescue_ramdisk',
                mutable=True,
-               deprecated_for_removal=True,
-               deprecated_reason=_('Replaced by rescue_ramdisk_by_arch which '
-                                   'provides more configuration options.'),
-               help=_('DEPRECATED: Glance ID, http:// or file:// URL of the '
+               help=_('Glance ID, http:// or file:// URL of the '
                       'initramfs of the default rescue image.')),
     cfg.DictOpt('rescue_kernel_by_arch',
                 default={},
@@ -262,8 +259,7 @@ opts = [
                help=_('Password hash algorithm to be used for the rescue '
                       'password.')),
     cfg.BoolOpt('require_rescue_password_hashed',
-                # TODO(TheJulia): Change this to True in Victoria.
-                default=False,
+                default=True,
                 mutable=True,
                 help=_('Option to cause the conductor to not fallback to '
                        'an un-hashed version of the rescue password, '
@@ -350,15 +346,36 @@ opts = [
                         'for multiple steps. If set to 0, this specific step '
                         'will not run during verification. ')),
     cfg.BoolOpt('automatic_lessee',
-                default=False,
+                default=True,
                 mutable=True,
-                help=_('If the conductor should record the Project ID '
-                       'indicated by Keystone for a requested deployment. '
-                       'Allows rights to be granted to directly access the '
-                       'deployed node as a lessee within the RBAC security '
-                       'model. The conductor does *not* record this value '
-                       'otherwise, and this information is not backfilled '
-                       'for prior instances which have been deployed.')),
+                help=_('Deprecated. If Ironic should set the node.lessee '
+                       'field at deployment. Use '
+                       '[\'conductor\']/automatic_lessee_source instead.'),
+                deprecated_for_removal=True),
+    cfg.StrOpt('automatic_lessee_source',
+               help=_('Source for Project ID the Ironic should '
+                      'record at deployment time in node.lessee field. If set '
+                      'to none, Ironic will not set a lessee field. '
+                      'If set to instance (default), uses Project ID '
+                      'indicated in instance metadata set by Nova or '
+                      'another external deployment service. '
+                      'If set to keystone, Ironic uses Project ID indicated '
+                      'by Keystone context. '),
+               choices=[
+                   (lessee_sources.INSTANCE, _(  # 'instance'
+                    'Populates node.lessee field using metadata from '
+                    'node.instance_info[\'project_id\'] at deployment '
+                    'time. Useful for Nova-fronted deployments.')),
+                   (lessee_sources.REQUEST, _(  # 'request'
+                    'Populates node.lessee field using metadata '
+                    'from request context. Only useful for direct '
+                    'deployment requests to Ironic; not those proxied '
+                    'via an external service like Nova.')),
+                   (lessee_sources.NONE, _(  # 'none'
+                    'Ironic will not populate the node.lessee field.'))
+               ],
+               default='instance',
+               mutable=True),
     cfg.IntOpt('max_concurrent_deploy',
                default=250,
                min=1,
@@ -423,6 +440,16 @@ opts = [
                       'seconds, or 30 minutes. If you need to wait longer '
                       'than the maximum value, we recommend exploring '
                       'hold steps.')),
+    cfg.ListOpt('disallowed_deployment_boot_modes',
+                item_type=types.String(
+                    choices=[
+                        (boot_modes.UEFI, _('UEFI boot mode')),
+                        (boot_modes.LEGACY_BIOS, _('Legacy BIOS boot mode'))],
+                ),
+                default=[],
+                mutable=True,
+                help=_("Specifies a list of boot modes that are not allowed "
+                       "during deployment. Eg: ['bios']")),
     cfg.BoolOpt('disable_deep_image_inspection',
                 default=False,
                 # Normally such an option would be mutable, but this is,
