@@ -20,9 +20,7 @@ from oslo_log import log
 from ironic.common import exception
 from ironic.common.i18n import _
 from ironic.common import neutron
-from ironic.common import states
 from ironic.drivers import base
-from ironic.drivers.modules import deploy_utils
 from ironic.drivers.modules.network import common
 
 LOG = log.getLogger(__name__)
@@ -35,22 +33,6 @@ class NeutronNetwork(common.NeutronVIFPortIDMixin,
                      base.NetworkInterface):
     """Neutron v2 network interface"""
 
-    def __init__(self):
-        failures = []
-        cleaning_net = CONF.neutron.cleaning_network
-        if not cleaning_net:
-            failures.append('cleaning_network')
-
-        provisioning_net = CONF.neutron.provisioning_network
-        if not provisioning_net:
-            failures.append('provisioning_network')
-
-        if failures:
-            raise exception.DriverLoadError(
-                driver=self.__class__.__name__,
-                reason=(_('The following [neutron] group configuration '
-                          'options are missing: %s') % ', '.join(failures)))
-
     def validate(self, task):
         """Validates the network interface.
 
@@ -59,17 +41,10 @@ class NeutronNetwork(common.NeutronVIFPortIDMixin,
             is invalid.
         :raises: MissingParameterValue, if some parameters are missing.
         """
+        # NOTE(TheJulia): These are the minimal networks needed for
+        # the neutron network interface to function.
         self.get_cleaning_network_uuid(task)
         self.get_provisioning_network_uuid(task)
-        node = task.node
-        if (node.provision_state == states.DEPLOYING
-            and node.driver_internal_info.get('is_whole_disk_image')
-            and deploy_utils.get_boot_option(node) == 'netboot'):
-            error_msg = (_('The node %s cannot perform "local" boot for '
-                           'whole disk image when node is using "neutron" '
-                           'network and is configured with "netboot" boot '
-                           'option.') % node.uuid)
-            raise exception.InvalidParameterValue(error_msg)
 
     def _add_network(self, task, network, security_groups, process):
         # If we have left over ports from a previous process, remove them
@@ -276,3 +251,33 @@ class NeutronNetwork(common.NeutronVIFPortIDMixin,
         """
         return self._remove_network(
             task, self.get_inspection_network_uuid(task), 'inspection')
+
+    def validate_servicing(self, task):
+        """Validates the network interface for servicing operation.
+
+        :param task: a TaskManager instance.
+        :raises: InvalidParameterValue, if the network interface configuration
+            is invalid.
+        :raises: MissingParameterValue, if some parameters are missing.
+        """
+        self.get_servicing_network_uuid(task)
+
+    def add_servicing_network(self, task):
+        """Create neutron ports for each port to boot the servicing ramdisk.
+
+        :param task: a TaskManager instance.
+        :returns: a dictionary in the form {port.uuid: neutron_port['id']}
+        """
+        return self._add_network(
+            task, self.get_servicing_network_uuid(task),
+            CONF.neutron.servicing_network_security_groups,
+            'servicing')
+
+    def remove_servicing_network(self, task):
+        """Deletes neutron port created for booting the servicing ramdisk.
+
+        :param task: a TaskManager instance.
+        :raises: NetworkError
+        """
+        return self._remove_network(
+            task, self.get_servicing_network_uuid(task), 'servicing')
