@@ -1,5 +1,6 @@
 # Copyright 2017 Red Hat, Inc.
 # All Rights Reserved.
+# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
@@ -14,22 +15,22 @@
 #    under the License.
 
 import collections
+import hashlib
 import os
 from urllib import parse as urlparse
 
 from oslo_log import log
 from oslo_utils import excutils
-from oslo_utils import importutils
 from oslo_utils import strutils
-import retrying
 import rfc3986
+import sushy
+import tenacity
 
 from ironic.common import exception
 from ironic.common.i18n import _
+from ironic.common import utils
 from ironic.conf import CONF
 
-sushy = importutils.try_import('sushy')
-
 LOG = log.getLogger(__name__)
 
 REQUIRED_PROPERTIES = {
@@ -77,6 +78,17 @@ COMMON_PROPERTIES = REQUIRED_PROPERTIES.copy()
 COMMON_PROPERTIES.update(OPTIONAL_PROPERTIES)
 
 
+# All available FIRMWARE COMPONENTS
+BIOS = 'bios'
+"BIOS Firmware Component"
+
+BMC = 'bmc'
+"BMC Firmware Component"
+
+FIRMWARE_COMPONENTS = [BIOS, BMC]
+"""Firmware Components available to update"""
+
+
 def parse_driver_info(node):
     """Parse the information required for Ironic to connect to Redfish.
 
@@ -95,7 +107,7 @@ def parse_driver_info(node):
                                                  'info': missing_info})
 
     # Validate the Redfish address
-    address = driver_info['redfish_address']
+    address = utils.wrap_ipv6(driver_info['redfish_address'])
     try:
         parsed = rfc3986.uri_reference(address)
     except TypeError:
@@ -107,8 +119,14 @@ def parse_driver_info(node):
     if not parsed.scheme or not parsed.authority:
         address = 'https://%s' % address
         parsed = rfc3986.uri_reference(address)
-    # TODO(vdrok): Workaround this check, in py3 we need to use validator class
-    if not parsed.is_valid(require_scheme=True, require_authority=True):
+    validator = rfc3986.validators.Validator().require_presence_of(
+        'scheme', 'host',
+    ).check_validity_of(
+        'scheme', 'userinfo', 'host', 'path', 'query', 'fragment',
+    )
+    try:
+        validator.validate(parsed)
+    except rfc3986.exceptions.RFC3986Exception:
         raise exception.InvalidParameterValue(
             _('Invalid Redfish address %(address)s set in '
               'driver_info/redfish_address on node %(node)s') %
@@ -180,59 +198,84 @@ def parse_driver_info(node):
 
 class SessionCache(object):
     """Cache of HTTP sessions credentials"""
-    AUTH_CLASSES = {}
-    if sushy:
-        AUTH_CLASSES.update(
-            basic=sushy.auth.BasicAuth,
-            session=sushy.auth.SessionAuth,
-            auto=sushy.auth.SessionOrBasicAuth
-        )
+
+    AUTH_CLASSES = dict(
+        basic=sushy.auth.BasicAuth,
+        session=sushy.auth.SessionAuth,
+        auto=sushy.auth.SessionOrBasicAuth
+    )
 
     _sessions = collections.OrderedDict()
 
     def __init__(self, driver_info):
+        # Hash the password in the data structure, so we can
+        # include it in the session key.
+        # NOTE(TheJulia): Multiplying the address by 4, to ensure
+        # we meet a minimum of 16 bytes for salt.
+        pw_hash = hashlib.pbkdf2_hmac(
+            'sha512',
+            driver_info.get('password').encode('utf-8'),
+            str(driver_info.get('address') * 4).encode('utf-8'), 40)
         self._driver_info = driver_info
+        # Assemble the session key and append the hashed password to it,
+        # which forces new sessions to be established when the saved password
+        # is changed, just like the username, or address.
         self._session_key = tuple(
             self._driver_info.get(key)
             for key in ('address', 'username', 'verify_ca')
-        )
+        ) + (pw_hash.hex(),)
 
     def __enter__(self):
         try:
             return self.__class__._sessions[self._session_key]
-
         except KeyError:
-            auth_type = self._driver_info['auth_type']
+            LOG.debug('A cached redfish session for Redfish endpoint '
+                      '%(endpoint)s was not detected, initiating a session.',
+                      {'endpoint': self._driver_info['address']})
 
-            auth_class = self.AUTH_CLASSES[auth_type]
+        auth_type = self._driver_info['auth_type']
 
-            authenticator = auth_class(
-                username=self._driver_info['username'],
-                password=self._driver_info['password']
-            )
+        auth_class = self.AUTH_CLASSES[auth_type]
 
-            sushy_params = {'verify': self._driver_info['verify_ca'],
-                            'auth': authenticator}
-            if 'root_prefix' in self._driver_info:
-                sushy_params['root_prefix'] = self._driver_info['root_prefix']
-            conn = sushy.Sushy(
-                self._driver_info['address'],
-                **sushy_params
-            )
+        authenticator = auth_class(
+            username=self._driver_info['username'],
+            password=self._driver_info['password']
+        )
 
-            if CONF.redfish.connection_cache_size:
-                self.__class__._sessions[self._session_key] = conn
+        sushy_params = {'verify': self._driver_info['verify_ca'],
+                        'auth': authenticator}
+        if 'root_prefix' in self._driver_info:
+            sushy_params['root_prefix'] = self._driver_info['root_prefix']
+        conn = sushy.Sushy(
+            self._driver_info['address'],
+            **sushy_params
+        )
 
-                if (len(self.__class__._sessions)
-                        > CONF.redfish.connection_cache_size):
-                    self._expire_oldest_session()
+        if CONF.redfish.connection_cache_size:
+            self.__class__._sessions[self._session_key] = conn
+            # Save a secure hash of the password into memory, so if we
+            # observe it change, we can detect the session is no longer valid.
 
-            return conn
+            if (len(self.__class__._sessions)
+                    > CONF.redfish.connection_cache_size):
+                self._expire_oldest_session()
+
+        return conn
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         # NOTE(etingof): perhaps this session token is no good
         if isinstance(exc_val, sushy.exceptions.ConnectionError):
             self.__class__._sessions.pop(self._session_key, None)
+        # NOTE(TheJulia): A hard access error has surfaced, we
+        # likely need to eliminate the session.
+        if isinstance(exc_val, sushy.exceptions.AccessError):
+            self.__class__._sessions.pop(self._session_key, None)
+        # NOTE(TheJulia): Something very bad has happened, such
+        # as the session is out of date, and refresh of the SessionService
+        # failed resulting in an AttributeError surfacing.
+        # https://storyboard.openstack.org/#!/story/2009719
+        if isinstance(exc_val, AttributeError):
+            self.__class__._sessions.pop(self._session_key, None)
 
     @classmethod
     def _expire_oldest_session(cls):
@@ -262,6 +305,40 @@ def get_update_service(node):
         raise exception.RedfishError(error=e)
 
 
+def get_event_service(node):
+    """Get a node's event service.
+
+    :param node: an Ironic node object.
+    :raises: RedfishConnectionError when it fails to connect to Redfish
+    :raises: RedfishError when the EventService is not registered in Redfish
+    """
+
+    try:
+        return _get_connection(node, lambda conn: conn.get_event_service())
+    except sushy.exceptions.MissingAttributeError as e:
+        LOG.error('The Redfish EventService was not found for '
+                  'node %(node)s. Error %(error)s',
+                  {'node': node.uuid, 'error': e})
+        raise exception.RedfishError(error=e)
+
+
+def get_first_controller(storage):
+    """Get the first storage controller from a storage object.
+
+    :param storage: a storage object
+    :returns: the first storage controller or None
+    """
+    try:
+        controllers = storage.controllers
+        if not controllers:
+            raise sushy.exceptions.MissingAttributeError
+        members = controllers.get_members()
+        return members[0] if members else None
+    except sushy.exceptions.MissingAttributeError:
+        controllers = getattr(storage, 'storage_controllers', None)
+        return controllers[0] if controllers else None
+
+
 def get_system(node):
     """Get a Redfish System that represents a node.
 
@@ -285,6 +362,47 @@ def get_system(node):
         raise exception.RedfishError(error=e)
 
 
+def get_system_collection(node):
+    """Get a Redfish System Collection that includes the node
+
+    :param node: an Ironic node object
+    :raises: RedfishConnectionError when it fails to connect to Redfish
+    :raises: RedfishError if the System is not registered in Redfish
+    """
+    driver_info = parse_driver_info(node)
+    system_id = driver_info['system_id']
+
+    try:
+        return _get_connection(
+            node,
+            lambda conn, system_id: conn.get_system_collection(),
+            system_id)
+    except sushy.exceptions.ResourceNotFoundError as e:
+        LOG.error('The Redfish Systems Collection "%(system)s" was not found'
+                  ' for node %(node)s. Error %(error)s',
+                  {'system': system_id or '<default>',
+                   'node': node.uuid, 'error': e})
+        raise exception.RedfishError(error=e)
+
+
+def get_task_monitor(node, uri):
+    """Get a TaskMonitor for a node.
+
+    :param node: an Ironic node object
+    :param uri: the URI of a TaskMonitor
+    :raises: RedfishConnectionError when it fails to connect to Redfish
+    :raises: RedfishError when the TaskMonitor is not available in Redfish
+    """
+
+    try:
+        return _get_connection(node, lambda conn: conn.get_task_monitor(uri))
+    except sushy.exceptions.ResourceNotFoundError as e:
+        LOG.error('The Redfish TaskMonitor "%(uri)s" was not found for '
+                  'node %(node)s. Error %(error)s',
+                  {'uri': uri, 'node': node.uuid, 'error': e})
+        raise exception.RedfishError(error=e)
+
+
 def _get_connection(node, lambda_fun, *args):
     """Get a Redfish connection to a node.
 
@@ -300,11 +418,12 @@ def _get_connection(node, lambda_fun, *args):
     """
     driver_info = parse_driver_info(node)
 
-    @retrying.retry(
-        retry_on_exception=(
-            lambda e: isinstance(e, exception.RedfishConnectionError)),
-        stop_max_attempt_number=CONF.redfish.connection_attempts,
-        wait_fixed=CONF.redfish.connection_retry_interval * 1000)
+    @tenacity.retry(
+        retry=tenacity.retry_if_exception_type(
+            exception.RedfishConnectionError),
+        stop=tenacity.stop_after_attempt(CONF.redfish.connection_attempts),
+        wait=tenacity.wait_fixed(CONF.redfish.connection_retry_interval),
+        reraise=True)
     def _get_cached_connection(lambda_fun, *args):
         try:
             with SessionCache(driver_info) as conn:
@@ -321,6 +440,23 @@ def _get_connection(node, lambda_fun, *args):
                          'auth_type': driver_info['auth_type'],
                          'node': node.uuid, 'error': e})
             raise exception.RedfishConnectionError(node=node.uuid, error=e)
+        except sushy.exceptions.AccessError as e:
+            LOG.warning('For node %(node)s, we received an authentication '
+                        'access error from address %(address)s with auth_type '
+                        '%(auth_type)s. The client will not be re-used upon '
+                        'the next re-attempt. Please ensure your using the '
+                        'correct credentials. Error: %(error)s',
+                        {'address': driver_info['address'],
+                         'auth_type': driver_info['auth_type'],
+                         'node': node.uuid, 'error': e})
+            raise exception.RedfishError(node=node.uuid, error=e)
+        except AttributeError as e:
+            LOG.warning('For node %(node)s, we received at AttributeError '
+                        'when attempting to utilize the client. A new '
+                        'client session shall be used upon the next attempt.'
+                        'Attribute Error: %(error)s',
+                        {'node': node.uuid, 'error': e})
+            raise exception.RedfishError(node=node.uuid, error=e)
 
     try:
         return _get_cached_connection(lambda_fun, *args)
@@ -330,3 +466,96 @@ def _get_connection(node, lambda_fun, *args):
                       'node %(node)s. Error: %(error)s',
                       {'address': driver_info['address'],
                        'node': node.uuid, 'error': e})
+
+
+def get_enabled_macs(task, system):
+    """Get information on MAC addresses of enabled ports using Redfish.
+
+    :param task: a TaskManager instance containing the node to act on.
+    :param system: a Redfish System object
+    :returns: a dictionary containing MAC addresses of enabled interfaces
+        in a {'mac': <state>} format, where <state> is a sushy constant
+    """
+
+    enabled_macs = {}
+    if (system.ethernet_interfaces
+            and system.ethernet_interfaces.summary):
+        macs = system.ethernet_interfaces.summary
+
+        # Identify ports for the NICs being in 'enabled' state
+        for nic_mac, nic_state in macs.items():
+            if nic_state != sushy.STATE_ENABLED:
+                continue
+            elif not nic_mac:
+                LOG.warning("Ignoring device for %(node)s as no MAC "
+                            "reported", {'node': task.node.uuid})
+                continue
+            enabled_macs[nic_mac] = nic_state
+
+    if not enabled_macs:
+        LOG.debug("No ethernet interface information is available "
+                  "for node %(node)s", {'node': task.node.uuid})
+    return enabled_macs
+
+
+def wait_until_get_system_ready(node):
+    """Wait until Redfish system is ready.
+
+    :param node: an Ironic node object
+    :raises: RedfishConnectionError on time out.
+    """
+    @tenacity.retry(
+        retry=tenacity.retry_if_exception_type(
+            exception.RedfishConnectionError),
+        stop=tenacity.stop_after_attempt(CONF.redfish.connection_attempts),
+        wait=tenacity.wait_fixed(CONF.redfish.connection_retry_interval),
+        reraise=True)
+    def _get_system(driver_info, system_id):
+        try:
+            with SessionCache(driver_info) as conn:
+                return conn.get_system(system_id)
+        except sushy.exceptions.BadRequestError as e:
+            err_msg = ("System is not ready for node %(node)s, with error"
+                       "%(error)s, so retrying it",
+                       {'node': node.uuid, 'error': e})
+            LOG.warning(err_msg)
+            raise exception.RedfishConnectionError(node=node.uuid, error=e)
+    driver_info = parse_driver_info(node)
+    system_id = driver_info['system_id']
+    return _get_system(driver_info, system_id)
+
+
+def get_manager(node, system, manager_id=None):
+    """Get a node's manager.
+
+    :param system: a Sushy system object
+    :param manager_id: the id of the manager
+    :return: a sushy Manager
+    :raises: RedfishError when the System doesn't have Managers associated
+    """
+
+    try:
+        sushy_manager = None
+        available_managers = system.managers
+        if available_managers:
+            if manager_id is None:
+                sushy_manager = available_managers[0]
+            else:
+                for manager in available_managers:
+                    if manager.identity == manager_id:
+                        sushy_manager = manager
+        if sushy_manager is None:
+            raise Exception("Couldn't find any Sushy Manager")
+        return sushy_manager
+    except sushy.exceptions.MissingAttributeError as e:
+        LOG.error('Redfish Managers for node %(node)s are not associated '
+                  'with system %(system)s. Error %(error)s',
+                  {'system': system.identity,
+                   'node': node.uuid, 'error': e})
+        raise exception.RedfishError(error=e)
+    except Exception as exc:
+        LOG.error('Redfish Manager was not found for '
+                  'node %(node)s under system %(system)s. Error %(error)s',
+                  {'system': system.identity,
+                   'node': node.uuid, 'error': exc})
+        raise exception.RedfishError(error=exc)
