@@ -373,7 +373,7 @@ class IronicImagesTestCase(base.TestCase):
                        autospec=True)
     def test_converted_size_estimate_default(self, image_info_mock):
         info = self.FakeImgInfo()
-        info.actual_size = 2
+        info.disk_size = 2
         info.virtual_size = 10 ** 10
         image_info_mock.return_value = info
         size = images.converted_size('path', estimate=True)
@@ -385,7 +385,7 @@ class IronicImagesTestCase(base.TestCase):
     def test_converted_size_estimate_custom(self, image_info_mock):
         CONF.set_override('raw_image_growth_factor', 3)
         info = self.FakeImgInfo()
-        info.actual_size = 2
+        info.disk_size = 2
         info.virtual_size = 10 ** 10
         image_info_mock.return_value = info
         size = images.converted_size('path', estimate=True)
@@ -397,7 +397,7 @@ class IronicImagesTestCase(base.TestCase):
     def test_converted_size_estimate_raw_smaller(self, image_info_mock):
         CONF.set_override('raw_image_growth_factor', 3)
         info = self.FakeImgInfo()
-        info.actual_size = 2
+        info.disk_size = 2
         info.virtual_size = 5
         image_info_mock.return_value = info
         size = images.converted_size('path', estimate=True)
@@ -716,13 +716,11 @@ class FsImageTestCase(base.TestCase):
         walk_mock.assert_called_once_with('tmpdir1')
         rmtree_mock.assert_called_once_with('tmpdir1')
 
-    @mock.patch.object(utils, 'mount', autospec=True)
-    def test__get_deploy_iso_files_fail_with_ExecutionError(
-            self, get_iso_files_mock):
-        get_iso_files_mock.side_effect = processutils.ProcessExecutionError
-        self.assertRaises(exception.ImageCreationFailed,
-                          images._get_deploy_iso_files,
-                          'path/to/deployiso', 'tmpdir1')
+    def test__get_deploy_iso_files_fail_with_ExecutionError(self):
+        self.assertRaisesRegex(exception.ImageCreationFailed,
+                               'No such file or directory',
+                               images._get_deploy_iso_files,
+                               'path/to/deployiso', 'tmpdir1')
 
     @mock.patch.object(shutil, 'rmtree', autospec=True)
     @mock.patch.object(images, '_create_root_fs', autospec=True)
