@@ -24,6 +24,7 @@ from unittest import mock
 import uuid
 
 from oslo_config import cfg
+from oslo_utils import timeutils
 from oslo_utils import uuidutils
 
 from ironic.common import exception
@@ -348,7 +349,7 @@ class TestUpdateImages(BaseTest):
         touch(self.master_path)
         href = 'http://awesomefreeimages.al/img999'
         self.img_info = {
-            'updated_at': datetime.datetime((datetime.datetime.utcnow().year
+            'updated_at': datetime.datetime((timeutils.utcnow().year
                                              + 1), 11, 15, 8, 12, 31)
         }
         res = image_cache._delete_master_path_if_stale(self.master_path, href,
@@ -585,7 +586,7 @@ class TestImageCacheCleanUp(base.TestCase):
 
         self.assertEqual(image_cache._cache_cleanup_list[0][1], Cache2)
 
-        # The order of caches with same prioirty is not deterministic.
+        # The order of caches with same priority is not deterministic.
         item_possibilities = [Cache1, Cache3]
         second_item_actual = image_cache._cache_cleanup_list[1][1]
         self.assertIn(second_item_actual, item_possibilities)
@@ -833,6 +834,43 @@ class TestFetchCleanup(base.TestCase):
         image_check.safety_check.assert_not_called()
         self.assertEqual(1, image_check.__str__.call_count)
 
+    @mock.patch.object(image_format_inspector, 'detect_file_format',
+                       autospec=True)
+    @mock.patch.object(images, 'image_show', autospec=True)
+    @mock.patch.object(os, 'remove', autospec=True)
+    @mock.patch.object(os.path, 'exists', autospec=True)
+    @mock.patch.object(images, 'converted_size', autospec=True)
+    @mock.patch.object(images, 'fetch', autospec=True)
+    @mock.patch.object(images, 'image_to_raw', autospec=True)
+    @mock.patch.object(image_cache, '_clean_up_caches', autospec=True)
+    def test__fetch_part_already_exists(
+            self, mock_clean, mock_raw, mock_fetch,
+            mock_size, mock_exists, mock_remove, mock_image_show,
+            mock_format_inspector):
+        image_check = mock.MagicMock()
+        image_check.__str__.side_effect = iter(['qcow2', 'raw'])
+        image_check.safety_check.return_value = True
+        mock_format_inspector.return_value = image_check
+        mock_exists.return_value = True
+        mock_size.return_value = 100
+        mock_image_show.return_value = {}
+        image_cache._fetch('fake', 'fake-uuid', '/foo/bar', force_raw=True,
+                           expected_format=None, expected_checksum='f00',
+                           expected_checksum_algo='sha256')
+        mock_fetch.assert_called_once_with('fake', 'fake-uuid',
+                                           '/foo/bar.part', force_raw=False,
+                                           checksum='f00',
+                                           checksum_algo='sha256')
+        mock_clean.assert_called_once_with('/foo', 100)
+        mock_raw.assert_called_once_with('fake-uuid', '/foo/bar',
+                                         '/foo/bar.part')
+        self.assertEqual(1, mock_exists.call_count)
+        self.assertEqual(1, mock_remove.call_count)
+        mock_image_show.assert_called_once_with('fake', 'fake-uuid')
+        mock_format_inspector.assert_called_once_with('/foo/bar.part')
+        image_check.safety_check.assert_called_once()
+        self.assertEqual(1, image_check.__str__.call_count)
+
     @mock.patch.object(os, 'rename', autospec=True)
     @mock.patch.object(image_format_inspector, 'detect_file_format',
                        autospec=True)
